import numpy as np
import pandas as pd
import random
import requests
import math
from IPython.display import clear_output
from IPython.core.display import HTML
try:
    import ipywidgets as widgets
    from ipywidgets import Button, Layout
except:    
    subprocess.check_call([sys.executable, '-m', 'pip', 'install', '--trusted-host', 'pypi.org', '--trusted-host', 'files.pythonhosted.org', 'ipywidgets'])
    clear_output(wait = True)
    import ipywidgets as widgets
    from ipywidgets import Button, Layout
try:
    import json
except:    
    subprocess.check_call([sys.executable, '-m', 'pip', 'install', '--trusted-host', 'pypi.org', '--trusted-host', 'files.pythonhosted.org', 'json'])
    clear_output(wait = True)
    import json
try:
    import requests
except:    
    subprocess.check_call([sys.executable, '-m', 'pip', 'install', '--trusted-host', 'pypi.org', '--trusted-host', 'files.pythonhosted.org', 'requests'])
    clear_output(wait = True)
    import requests

class concord_sym:

    def print_style(text, color='black', weight = 'bold', size = '10pt'):
        source = "<p style='color: "+color+"; font-weight:"+weight+"; font-size:"+size+"'>"+text+"</p>"
        return display(HTML(source))
    
    identifiers = ['ISIN', 'CUSIP', 'SEDOL', 'Ticker', 'Bloomberg FIGI', 'Bloomberg Ticker', 'Bank Identification Code', 'Edgar Central Index Keys', 'Central Registration Depository', 'Dun & Bradstreet DUNS', 'Employer Identification Number', 'Fitch Ratings Identifier', 'Legal Entity Identifier', 'Moodys Ratings Identifier', 'S&P Ratings Identifier', 'Valoren', 'German Securities Identification']
    
    reg_ids=['ISIN','CUSIP','SEDOL','Ticker','Bloomberg FIGI', 'Bloomberg Ticker']
    
    advanced_dict = {'Bank Identification Code' : 'BIC', 'Edgar Central Index Keys' : 'CIK', 'Central Registration Depository' : 'CRD', 'Dun & Bradstreet DUNS' : 'DUNS', 'Employer Identification Number' : 'EIN', 'Fitch Ratings Identifier' : 'FITCH', 'Legal Entity Identifier' : 'LEI', 'Moodys Ratings Identifier' : 'MD', 'S&P Ratings Identifier' : 'SPR', 'Valoren' : 'VALOREN', 'German Securities Identification' : 'WKN'}
    
    def api_creds(self):
        self.user = input('What is your FactSet API Username? ')
        self.passw = input('What is your FactSet API password? ')
        self.authorization = (self.user, self.passw)
        
    def import_file(self):
        self.file_path = input('Enter the directory of your file: ')
        if self.file_path[-3:] == 'csv':
            self.sym_table = pd.read_csv(self.file_path)
            self.sym_table = self.sym_table.replace(np.nan, "")
            print()
            
            self.field_names = widgets.SelectMultiple(options = self.sym_table.columns.to_list(), layout=Layout(height='120px'))
            concord_sym.print_style('Select the fields in your file which contain market IDs (hold down CTRL to select multiple)')
            display(self.field_names)
            print()
            print('_______________________________________________________________________')
            print()
            self.concord_names = widgets.Select(options = self.sym_table.columns.to_list())
            concord_sym.print_style('Select the field in your file which contains company names: ')
            display(self.concord_names)
        else:
            concord_sym.print_style('The input file must be in csv format.', color = 'red')
        
    def market_id_type(self):
        if self.field_names.value == ():
            concord_sym.print_style('You did not specify any market ID fields in the above step. You may do so now and rerun this cell; otherwise mappings will only be run using the Concordance API on your Company Name field.', color='red')
        else:
            for i in self.field_names.value:
                print(i)
                exec('self.'+i + " = widgets.RadioButtons(options=concord_sym.identifiers)")
                exec('display(self.'+i+')')
                print()
            
    def mappings(self):
        self.id_dict = {}
        for a in self.field_names.value:
            exec('self.id_dict[a] = self.'+a+'.value')            
        self.id_dict[self.concord_names.value]='Company Name'

        for key, value in self.id_dict.items():
            if value == 'Moodys Ratings Identifier':
                self.sym_table[key]=self.sym_table[key].apply('{:0>12}'.format)        
                self.sym_table.loc[self.sym_table[key] == '000000000nan', key] = 9999999999
                self.sym_table[key]=self.sym_table[key].astype('str').str[:-2]
            else:
                col_list = []
                for val in self.sym_table[key]:
                    if type(val) == float:
                        col_list += [str(int(val))]
                    else:
                        col_list += [str(val)]
                self.sym_table[key] = col_list
                
        url = 'https://api.factset.com/content/symbology/v2/factset-advanced'
        self.symbology_df = pd.DataFrame({})
        for i in self.id_dict:
            if self.id_dict[i] != 'Company Name':
                try:
                    if self.id_dict[i] in concord_sym.reg_ids:
                        params = {"ids" : self.sym_table[i].to_list()}
                        response = requests.get(url=url, params = params, auth = self.authorization)
                        exec(i + "_df = pd.DataFrame(response.json()['data'])[['fsymEntityId', 'requestId']]")
                        exec(i+"_df.columns = ['"+i+"_fsymEntityId', '"+i+"_requestId']")
                        exec(i+"_df = self.sym_table[['"+i+"']].merge("+i+"_df, how = 'left', left_on='"+i+"', right_on='"+i+"_requestId')")
                        exec(i+"_df = "+i+"_df.drop('"+i+"', axis =1)")
                        exec("self.symbology_df="+i+"_df.join(self.symbology_df)")
                    else:
                        params = {"ids" : self.sym_table[i].to_list(), "symbolType" : concord_sym.advanced_dict[self.id_dict[i]]}
                        response = requests.get(url=url, params = params, auth = self.authorization)
                        exec(i + "_df = pd.DataFrame(response.json()['data'])[['fsymEntityId', 'requestId']]")
                        exec(i+"_df.columns = ['"+i+"_fsymEntityId', '"+i+"_requestId']")
                        exec(i+"_df['"+i+"_requestId'] = "+i+"_df['"+i+"_requestId'].astype('str')")
                        self.sym_table[i] = self.sym_table[i].astype('str')
                        exec(i+"_df = self.sym_table[['"+i+"']].merge("+i+"_df, how='left', left_on='"+i+"', right_on='"+i+"_requestId')")
                        exec(i+"_df = "+i+"_df.drop('"+i+"', axis =1)")
                        exec("self.symbology_df="+i+"_df.join(self.symbology_df)")
                except ValueError:
                    concord_sym.print_style('ERROR: The Symbology API returned no response. Please ensure your API credentials have access to the Symbology API.', color='red')
                    continue
            else:
                try:
                    concord_url = "https://api.factset.com/content/factset-concordance/v2/entity-match"
                    headers = {'accept': 'application/json;charset=utf-8','Content-Type': 'application/json;charset=utf-8'}
                    raw_df = pd.DataFrame()
                    itr = math.ceil(len(self.sym_table[i])/25)
                    n=0
                    start = 0
                    
                    while n<itr:
                        req = {"input":[]}
                        table_chunk = self.sym_table[i][start:start+25]
                        for name in table_chunk:
                            req["input"].append({"name":name})
                        conc_response = requests.post(url=concord_url, headers = headers, data = json.dumps(req), auth = self.authorization)
                        try:
                            raw_df = raw_df.append(pd.DataFrame(conc_response.json()['data']))
                        except:
                            pass
                        n+=1
                        start+=25
                        continue
                    if not(raw_df.empty):
                        conf_df = raw_df[raw_df['confidenceScore']==1]
                        decision = {i+"_fsymEntityId":[],i+"_requestId":[]}
                        for name in self.sym_table[i]:
                            decision[i+"_requestId"].append(name)
                            if name in conf_df['clientName'].values:
                                decision[i+"_fsymEntityId"].append(conf_df['entityId'][conf_df['clientName']==name].values[0])
                            else:
                                decision[i+"_fsymEntityId"].append('')
                        exec(i+"_df = pd.DataFrame(decision)")
                        exec("self.symbology_df="+i+"_df.join(self.symbology_df)")
                        clear_output(wait = True)
                    elif not(self.symbology_df.empty):
                        exec(i+"_df = pd.DataFrame({'"+i+"_fsymEntityId':['']*len(self.sym_table), '"+i+"_requestId':self.sym_table[i]})")
                        exec("self.symbology_df="+i+"_df.join(self.symbology_df)")
                        concord_sym.print_style('ERROR: The Concordance API returned no results. Mappings will be made only on the market IDs you provided. Please ensure that the Company Name field in your file is correct and that your API credentials have access to the Concordance API.', color='red')
                    else:
                        self.symbology_df = pd.DataFrame()
                        concord_sym.print_style('ERROR: The Concordance API returned no response. Please ensure your API credentials have access to the Concordance API.', color='red')
                except:
                    concord_sym.print_style('ERROR: The Concordance API returned no response. Please ensure your API credentials have access to the Concordance API.', color='red')
                    break
        if not(self.symbology_df.empty):
            self.symbology_df=self.symbology_df.where(pd.notnull(self.symbology_df), None)
            
            mapped_dict = {'ordered_row':[],'factset_entity_id':[],'request_id':[],'request_id_type':[]}
            
            ordered_ids = []
            for a in concord_sym.identifiers:
                if a in self.id_dict.values():
                    for key, value in self.id_dict.items():
                        if value == a:
                            ordered_ids.append(key)
            for key, value in self.id_dict.items():
                        if value == 'Company Name':
                            ordered_ids.append(key)  
            
            for i in ordered_ids:
                for index, row in self.symbology_df.iterrows():
                    if index not in mapped_dict['ordered_row'] and row[i+"_fsymEntityId"] != None and row[i+"_fsymEntityId"] != "":
                        mapped_dict['ordered_row'].append(index)
                        mapped_dict['factset_entity_id'].append(row[i+"_fsymEntityId"])
                        mapped_dict['request_id'].append(row[i+"_requestId"])
                        mapped_dict['request_id_type'].append(self.id_dict[i])
                        
            self.mapped_df = pd.DataFrame(mapped_dict).sort_values(by='ordered_row').reset_index(drop=True).drop(columns=['ordered_row'])
            
            self.unmapped_df=pd.DataFrame()

            for index, row in self.symbology_df.iterrows():
                if index not in mapped_dict['ordered_row']:
                    self.unmapped_df = self.unmapped_df.append(row)
            folder_path=input('Please specify the folder path where you would like your mapping files to be saved (e.g. C:\\\\Users\\\\jsmith\\\\my_symbology_folder) ')
            if folder_path == '':
                self.mapped_df.to_csv('symbology_mappings.csv', index=False)
                self.unmapped_df.to_csv('unmapped_entities.csv', index=False)
                self.symbology_df.to_csv('raw_mapping_output.csv', index=False)
            else:
                if folder_path[-1:] == '\\':
                    self.mapped_df.to_csv(folder_path+'symbology_mappings.csv', index=False)
                    self.unmapped_df.to_csv(folder_path+'unmapped_entities.csv', index=False)
                    self.symbology_df.to_csv(folder_path+'raw_mapping_output.csv', index=False)
                else:
                    self.mapped_df.to_csv(folder_path+'\\symbology_mappings.csv', index=False)
                    self.unmapped_df.to_csv(folder_path+'\\unmapped_entities.csv', index=False)
                    self.symbology_df.to_csv(folder_path+'\\raw_mapping_output.csv', index=False)
            
            concord_sym.print_style('3 Files have been saved to your folder: ', color = 'blue')
            print()
            concord_sym.print_style('symbology_mapping.csv')
            print('This is the final output of mappings made on the prioritized provided identifier')
            display(self.mapped_df.head(3))
            print()
            print('_______________________________________________________________________')
            print()
            concord_sym.print_style('unmapped_entities.csv')
            print('These are the rows from the file you provided that were NOT mapped')
            display(self.unmapped_df.head(3))
            print()
            print('_______________________________________________________________________')
            print()
            concord_sym.print_style('raw_mapping_output.csv')
            print('These are the raw mappings. For every row, this provides the FactSet ID mapped from each specified identifier')
            display(self.symbology_df.head(3))
        
        
    
    def test_display(self):
        display(self.sym_table)